import torch
from torch import Tensor
import torch.nn as nn

from .PositionalEncoding import PositionalEncoding
from .RelativePositionalRepresentations import CustomTransformerEncoder
from .rpr import TransformerEncoderLayerRPR, TransformerEncoderRPR
from .progress import LearningProgress


class MORTM(nn.Module):
    token_dict = {
        'SHIFT': [1, 6],
        'START': [7, 38],
        'PITCH': [39, 166],
        'DURATION': [167, 266]
    }

    def __init__(self, vocab_size, progress: LearningProgress, trans_layer=6, num_heads=8, d_model=512,
                 dim_feedforward=1024, dropout=0.1,
                 position_length=2048, use_rpr=True, use_decoder=False):
        super(MORTM, self).__init__()

        if use_decoder:
            print("Use decoder mode!")

        self.progress = progress
        self.trans_layer = trans_layer
        self.num_heads = num_heads
        self.d_model = d_model
        self.dim_feedforward = dim_feedforward
        self.dropout = dropout

        #位置エンコーディングを作成
        #self.positional: LearnablePositionalEncoding = LearnablePositionalEncoding(self.d_model, progress, dropout, position_length).to(self.progress.get_device())
        self.positional: PositionalEncoding = PositionalEncoding(self.d_model, progress, dropout, position_length).to(self.progress.get_device())
        #Transformerの設定
        if not use_rpr:
            self.transformer: nn.Transformer = nn.Transformer(d_model=self.d_model, nhead=num_heads,  #各種パラメーターの設計
                                                              num_encoder_layers=self.trans_layer,
                                                              num_decoder_layers= self.trans_layer if use_decoder else 0, #デコーダーは使うときと使わない時がある。
                                                              dropout=self.dropout, dim_feedforward=dim_feedforward,
                                                              custom_decoder=None if use_decoder else DummyDecoder()
                                                              ).to(self.progress.get_device())
        else:
            encoder_norm = nn.LayerNorm(self.d_model)
            encoder_layer = TransformerEncoderLayerRPR(self.d_model, self.num_heads, self.dim_feedforward, self.dropout, er_len=position_length)
            encoder = TransformerEncoderRPR(encoder_layer, self.trans_layer, encoder_norm)
            self.transformer = nn.Transformer(
                d_model=self.d_model, nhead=self.num_heads, num_encoder_layers=self.trans_layer,
                num_decoder_layers=self.trans_layer if use_decoder else 0,
                dropout=self.dropout, # activation=self.ff_activ,
                dim_feedforward=self.dim_feedforward,
                custom_decoder=None if use_decoder else DummyDecoder(),
                custom_encoder=encoder
            ).to(device=progress.get_device())

            print("Use RPR Transformer")
        print(f"Input Vocab Size:{vocab_size}")
        self.Wout: nn.Linear = nn.Linear(self.d_model, vocab_size).to(self.progress.get_device())

        self.embedding: nn.Embedding = nn.Embedding(vocab_size, self.d_model).to(self.progress.get_device())
        self.softmax: nn.Softmax = nn.Softmax(dim=-1).to(self.progress.get_device())

    def forward(self, inputs_seq, tgt_seq=None, src_mask=None, tgt_mask=None, input_padding_mask=None, tgt_padding_mask=None):
        if tgt_seq is None or src_mask is None:
            mask = self.transformer.generate_square_subsequent_mask(inputs_seq.shape[1]).to(self.progress.get_device())
        elif src_mask is not None:
            mask = src_mask
        else:
            mask = None

        inputs_em: Tensor = self.embedding(inputs_seq)
        inputs_em = inputs_em.permute(1, 0, 2)

        inputs_pos: Tensor = self.positional(inputs_em)

        if tgt_seq is not None:
            tgt_e = self.embedding(tgt_seq)
            tgt_e = tgt_e.permute(1, 0, 2)
            tgt_p = self.positional(tgt_e)
        else:
            tgt_p = inputs_pos

        out: Tensor = self.transformer(inputs_pos, tgt_p, src_mask=mask, tgt_mask=tgt_mask,
                                       src_key_padding_mask=input_padding_mask, tgt_key_padding_mask=tgt_padding_mask)

        out.permute(1, 0, 2)

        score:Tensor = self.Wout(out)
        return score.to(self.progress.get_device())

    def top_k_sampling_length_encoder(self, input_sequence, temperature=1.0, top_k=3, max_length=100):
        self.eval()
        """
        MORTMモデルにトップKサンプリングと温度シーケンスを実装する関数

        Args:
            input_sequence: 1次元の入力シーケンス (List[int] or torch.Tensor)。
            temperature: 温度パラメータ。デフォルトは1.0。
            top_k: サンプリングする上位K個のトークンの数。デフォルトは3。
            max_length: 生成するシーケンスの最大長さ。デフォルトは100。

        Returns:
            1次元の生成されたシーケンス (torch.Tensor)。
        """

        # 入力シーケンスをtorch.tensorに変換
        if not isinstance(input_sequence, torch.Tensor):
            input_sequence = torch.tensor(input_sequence, dtype=torch.long, device=self.progress.get_device())

        # 入力シーケンスの長さを取得
        input_length = input_sequence.size(0)

        # 生成されたシーケンスを格納するリスト
        generated_sequence = input_sequence.tolist()

        # 生成をループ
        for i in range(max_length):

            print(f"\r Generating...{i / max_length * 100}%", end="")

            # モデルに渡すための入力の準備 (2次元に変換)
            input_tensor = input_sequence.unsqueeze(0)  # (1, sequence_length)
            #print(f"I{i} input_tensor")

            # モデルに入力して次のトークンのスコアを取得 (3次元で返ってくる)
            with torch.no_grad():
                mask = self.transformer.generate_square_subsequent_mask(input_tensor.shape[1]).to(
                    self.progress.get_device())
                #print(input_tensor.shape)
                scores = self(input_tensor)  # (1, sequence_length, vocab_size)
                #print(f"SCORE: {scores.shape}")

            # 最新のトークンのスコアを取得 (最後のトークンに対するスコア)
            logits = scores[:, -1, :]  # (1, vocab_size)

            # 温度の適用
            logits = logits / temperature
            logits = logits[-1, :]

            # ソフトマックスを適用して確率を取得
            probs = self.softmax(logits)  # (vocab_size)

            # トップKの確率でトークンをフィルタリング
            sorted_probs, sorted_indices = torch.topk(probs, top_k)
            #print(sorted_probs, sorted_indices)

            # 再度正規化
            sorted_probs = sorted_probs / sorted_probs.sum(dim=-1, keepdim=True)

            # トークンをサンプリング
            sampled_index = torch.multinomial(sorted_probs, 1).item()  # サンプリングされたインデックスを取得

            # ソートされたインデックスから元のインデックスに変換
            next_token = sorted_indices[sampled_index].item()

#            print(next_token)

            # シーケンスにトークンを追加
            generated_sequence.append(next_token)

            # 次のステップの入力として準備
            input_sequence = torch.tensor(generated_sequence, dtype=torch.long, device=self.progress.get_device())

        return input_sequence

    def top_k_sampling_length_decoder(self, input_sequence, temperature=1.0, top_k=3, max_length=100):
        self.eval()
        """
        MORTMモデルにトップKサンプリングと温度シーケンスを実装する関数

        Args:
            input_sequence: 1次元の入力シーケンス (List[int] or torch.Tensor)。
            temperature: 温度パラメータ。デフォルトは1.0。
            top_k: サンプリングする上位K個のトークンの数。デフォルトは3。
            max_length: 生成するシーケンスの最大長さ。デフォルトは100。

        Returns:
            1次元の生成されたシーケンス (torch.Tensor)。
        """

        # 入力シーケンスをtorch.tensorに変換
        if not isinstance(input_sequence, torch.Tensor):
            input_sequence = torch.tensor(input_sequence, dtype=torch.long, device=self.progress.get_device())

        # 入力シーケンスの長さを取得
        input_length = input_sequence.size(0)

        # 生成されたシーケンスを格納するリスト
        generated_sequence = [11]

        # 生成をループ
        for i in range(max_length):
            # モデルに渡すための入力の準備 (2次元に変換)
            tgt = torch.tensor(generated_sequence, device=self.progress.get_device()).unsqueeze(0)  # (1, sequence_length)
            src = input_sequence.unsqueeze(0)
            #print(f"I{i} input_tensor")

            # モデルに入力して次のトークンのスコアを取得 (3次元で返ってくる)
            with torch.no_grad():
                mask = self.transformer.generate_square_subsequent_mask(tgt.shape[1]).to(
                    self.progress.get_device())
                print(src.shape)
                scores = self(src, tgt_seq=tgt)  # (1, sequence_length, vocab_size)
                print(f"SCORE: {scores.shape}")

            # 最新のトークンのスコアを取得 (最後のトークンに対するスコア)
            logits = scores[:, -1, :]  # (1, vocab_size)

            # 温度の適用
            logits = logits / temperature
            logits = logits[-1, :]

            # ソフトマックスを適用して確率を取得
            probs = self.softmax(logits)  # (vocab_size)

            # トップKの確率でトークンをフィルタリング
            sorted_probs, sorted_indices = torch.topk(probs, top_k)
            #print(sorted_probs, sorted_indices)

            # 再度正規化
            sorted_probs = sorted_probs / sorted_probs.sum(dim=-1, keepdim=True)

            # トークンをサンプリング
            sampled_index = torch.multinomial(sorted_probs, 1).item()  # サンプリングされたインデックスを取得

            # ソートされたインデックスから元のインデックスに変換
            next_token = sorted_indices[sampled_index].item()

            #            print(next_token)

            # シーケンスにトークンを追加
            generated_sequence.append(next_token)

            # 次のステップの入力として準備
#            input_sequence = torch.tensor(generated_sequence, dtype=torch.long, device=self.progress.get_device())
        input_sequence = torch.cat((input_sequence, torch.tensor(generated_sequence, device=self.progress.get_device())))
        return input_sequence

    def top_p_sampling_length(self, input_seq, p=0.8, max_length=20, temperature=0.2):
        self.eval()
        if not isinstance(input_seq, torch.Tensor):
            input_seq = torch.tensor(input_seq, dtype=torch.long, device=self.progress.get_device())

        generated = input_seq.tolist()
        for _ in range(max_length):
            logits= self(input_seq)
            token = self.top_p_sampling(logits, p=p, temperature=temperature)
            generated.append(token)

        input_seq = torch.tensor(generated, dtype=torch.long, device=self.progress.get_device())
        return input_seq

    def top_p_sampling(self, logits, p=0.9, temperature=1.0)-> int:

        logits = logits / temperature
        # logitsをソフトマックスで確率分布に変換
        probs = self.softmax(logits, dim=-1)
        # 確率の降順に並べ替え、そのインデックスを取得
        sorted_probs, sorted_indices = torch.sort(probs, descending=True)

        # 累積確率を計算
        cumulative_probs = torch.cumsum(sorted_probs, dim=-1)

        # 累積確率がpを超えるインデックスを取得
        cutoff_index = torch.where(cumulative_probs > p)[0][0]

        # 上位pに入らないトークンの確率を0にする
        sorted_probs[cutoff_index + 1:] = 0

        # 確率を再正規化
        sorted_probs /= torch.sum(sorted_probs)

        # トークンをサンプリング
        sampled_index = torch.multinomial(sorted_probs, 1)

        # インデックスを元の順序に戻す
        return sorted_indices[sampled_index].item()


class DummyDecoder(nn.Module):
    def __init__(self):
        super(DummyDecoder, self).__init__()

    def forward(self, tgt, memory, tgt_mask, memory_mask, tgt_key_padding_mask, memory_key_padding_mask, **kwargs):
        return memory



from pretty_midi.pretty_midi import Note

from abc import abstractmethod


def ct_time_to_beat(time: float, tempo: int) -> int:
    b4 = 60 / tempo
    b8 = b4 / 2
    b16 = b8 / 2
    b32 = b16 / 2
    b64 = b32 / 2

    beat, sub = calc_time_to_beat(time, b64)

    return beat


def ct_beat_to_time(beat: float, tempo: int) -> float:
    b4 = 60 / tempo
    b8 = b4 / 2
    b16 = b8 / 2
    b32 = b16 / 2
    b64 = b32 / 2
    return beat * b64


def calc_time_to_beat(time, beat_time) -> (int, int):
    main_beat: int = time // beat_time
    sub_time: int = time % beat_time
    return main_beat, sub_time


class Token:
    def __init__(self, token_type: str, convert_type: int):
        self.token_type = token_type
        self.token_position = 0
        self.convert_type = convert_type

    @abstractmethod
    def get_token(self, back_notes: Note, note: Note, tempo: int) -> int | str | None:
        pass

    @abstractmethod
    def get_range(self) -> int:
        pass

    @abstractmethod
    def de_convert(self, number: int | str, back_note:Note, note: Note, tempo: int):
        pass

    @abstractmethod
    def set_tokens(self, tokens: dict):
        pass

    def __call__(self, back_notes: Note = None, note: Note = None, token: str = None, tempo: int = 120, *args,
                 **kwargs):
        if self.convert_type == 0:
            symbol: int | str | None = self.get_token(back_notes, note, tempo)
            if symbol == -999 or symbol is None:
                my_token = None
                pass
            else:
                my_token = f"{self.token_type}_{symbol}"
        else:
            if token is None:
                return None
            split = token.split("_")
            if split[0] == self.token_type:
                try:
                    symbol = int(float(split[-1]))
                except (ValueError, TypeError):
                    symbol = split[-1]
                self.de_convert(symbol, back_notes, note, tempo)
                return split[0]
            else:
                return None

        return my_token


class MeasureToken(Token):

    def get_range(self) -> int:
        return 1

    def de_convert(self, number: int, b, n, tempo: int):
        pass

    def set_tokens(self, tokens: dict):
        tokens[f'm_start'] = len(tokens)

    def get_token(self, back_notes: Note, note: Note, tempo: int) -> int or None or str:
        measure1 = 60 / tempo * 4
        if back_notes is not None:
            note_measure = note.start // measure1
            back_note_measure = back_notes.start // measure1
            if note_measure > back_note_measure:
                return "start"
            else:
                return None
        else:
            return "start"


class StartRE(Token):

    def set_tokens(self, tokens: dict):
        max_length = 192
        tokens_length = len(tokens)
        for i in range(max_length + 1):
            tokens[f's_{i}'] = tokens_length + i

    def de_convert(self, number: int, back_note, note: Note, tempo):
        shift = ct_beat_to_time(number, tempo)
        note.start = shift if back_note is None else shift + back_note.start

    def get_range(self) -> int:
        return 193

    def get_token(self, back_notes: Note, note: Note, tempo) -> int:
        now_start = ct_time_to_beat(note.start, tempo)
        if back_notes is not None:
            back_start = ct_time_to_beat(back_notes.start, tempo)
        else:
            back_start = 0
        shift = int(now_start - back_start)

        if shift < 0:
            print("WHATS!?!?!?!?!?")

        if shift > 192:
            shift = 128 + shift % 64
        if back_notes is None:
            shift = shift % 64
        return shift


class Pitch(Token):

    def set_tokens(self, tokens: dict):
        max_length = 127
        tokens_length = len(tokens)
        for i in range(max_length + 1):
            tokens[f'p_{i}'] = tokens_length + i

    def de_convert(self, number: int, back_note, note: Note, tempo):
        note.pitch = number

    def get_range(self) -> int:
        return 129

    def get_token(self, back_notes: Note, note: Note, tempo) -> int:
        p: int = note.pitch
        return p


class Velocity(Token):

    def set_tokens(self, tokens: dict):
        max_length = 127
        tokens_length = len(tokens)
        for i in range(max_length + 1):
            tokens[f'v_{i}'] = tokens_length + i

    def de_convert(self, number: int, b, n: Note, tempo):
        n.velocity = number

    def get_token(self, back_notes: Note, note: Note, tempo) -> int:
        v: int = note.velocity
        return v

    def get_range(self) -> int:
        return 128


class Duration(Token):

    def set_tokens(self, tokens: dict):
        max_length = 192
        tokens_length = len(tokens)
        for i in range(max_length + 1):
            tokens[f'd_{i}'] = tokens_length + i

    def de_convert(self, number: int, back_note: Note, note: Note, tempo):
        duration = ct_beat_to_time(number, tempo)
        note.end = note.start + duration

    def get_token(self, back_notes: Note, note: Note, tempo) -> int:
        start = ct_time_to_beat(note.start, tempo)
        end = ct_time_to_beat(note.end, tempo)
        d = int(max(abs(end - start), 1))

        if 192 < d:
            d = 191

        return d

    def get_range(self) -> int:
        return 192


class Start(Token):

    def de_convert(self, number: int, tempo):
        return ct_beat_to_time(number, tempo)

    def get_token(self, back_notes: Note, note: Note, tempo) -> int:
        s = ct_time_to_beat(note.start, tempo)
        return s % 64

    def get_range(self) -> int:
        return 64


class Shift(Token):

    def de_convert(self, number: int, tempo):
        b4 = 60 / tempo
        b8 = b4 / 2
        b16 = b8 / 2
        b32 = b16 / 2
        b64 = b32 / 2
        return b64 * 64 * number

    def get_token(self, back_notes: Note, note: Note, tempo) -> int:
        if back_notes is None:
            return 0
        else:
            back_start = ct_time_to_beat(back_notes.start, tempo)
            note_start = ct_time_to_beat(note.start, tempo)

            shift = int(abs((back_start // 64) - (note_start // 64)))

            if shift > 3:
                shift = 3

        return shift

    def get_range(self) -> int:
        return 4



from abc import abstractmethod
import torch



class LearningProgress:

    @abstractmethod
    def step_optimizer(self, optimizer, model, accumulation_steps, **kwargs):
        pass

    @abstractmethod
    def get_device(self):
        pass


class _DefaultLearningProgress(LearningProgress):

    def get_device(self):
        if torch.cuda.is_available():
            return torch.device('cuda')
        else:
            return torch.device('cpu')
        pass

    def step_optimizer(self, optimizer, model, accumulation_steps, **kwargs):
        norm = self.get_gradient_norm(model)
        if not (0.1 < norm < 3.0):
            print(
                f"\033[31m 警告\033[0m：NORMが既定値から逸脱しています。学習率、またはバッチサイズを調整してください。({norm:.4f})")


        torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=1.5)

        optimizer.step()  # オプティマイザを更新
        optimizer.zero_grad()
        pass

    def get_gradient_norm(self, model):
        total_norm = 0
        # モデルのすべてのパラメータについて、勾配がNoneでないものを対象にする
        for param in model.parameters():
            if param.grad is not None:
                param_norm = param.grad.detach().data.norm(2)  # L2ノルムを計算
                total_norm += param_norm.item() ** 2  # 勾配ノルムの2乗を足す
        total_norm = total_norm ** 0.5  # 最終的に平方根をとってL2ノルムを計算
        return total_norm



import json
from typing import List
from numpy import ndarray

import re
from . import constants
from .aya_node import Token, Start,Shift,Pitch,Velocity,Duration, StartRE, MeasureToken

'''旋律トークン'''
PITCH_TYPE = 'p'
VELOCITY_TYPE = 'v'
DURATION_TYPE = 'd'

'''指示トークン'''
START_TYPE = 's'
SHIFT_TYPE = 'h'
MEASURE_TYPE = 'm'

'''変換の向き'''
TO_TOKEN = 0
TO_MUSIC = 1

def get_token_converter(convert: int) -> List[Token]:
    register: List[Token] = list()

    register.append(MeasureToken(MEASURE_TYPE, convert))
    register.append(StartRE( START_TYPE, convert))
    register.append(Pitch( PITCH_TYPE, convert))
    #register.append(Velocity(tempo, VELOCITY_TYPE))
    register.append(Duration(DURATION_TYPE, convert))

    return register


class Tokenizer:
    def __init__(self, token: List[Token], load_data: str = None):
        if load_data is None:
            # 特殊トークン
            self.special_token_position = 3
            self.token_list = token
            self.tokens: dict = dict()
            self.tokens[constants.PADDING_TOKEN] = 0
            self.tokens[constants.START_SEQ_TOKEN] = 1
            self.tokens[constants.END_SEQ_TOKEN] = 2
            for t in token:
                t.set_tokens(self.tokens)

            self.token_max: dict = self._init_mx_dict(len(self.tokens))
            self.is_converter = False
        else:
            self.is_converter = True
            self.token_list = token
            with open(load_data, 'r') as file:
                self.tokens: dict = json.load(file)
                self.rev_tokens: dict = {v: k for k, v in self.tokens.items()}

    def _init_mx_dict(self, mx) -> dict:
        my_dict = dict()
        for i in range(0, mx):
            my_dict[i] = 0
        return my_dict

    def rev_get(self, a):
        return self.rev_tokens[a]

    def get(self, token: str):
        if token is not None:
            if token not in self.tokens and "<" not in token:
                sp = token.split('_')
                length = self.get_length(sp[0])

                self.tokens[token] = length
            if not self.is_converter:
                self.token_max[self.tokens[token]] += 1
            return self.tokens[token]

    def get_length(self, token_type: str):
        for t in self.token_list:
            if t.token_type == token_type:
                p = t.token_position
                t.token_position += 1
                return p
        pass

    def save(self, save_directory):
        json_string = json.dumps(self.tokens)
        with open(save_directory + "/vocab_list.json", 'w') as file:
            file.write(json_string)

        json_s = json.dumps(self.token_max)
        with open(save_directory + "/vocab_max.json", 'w') as file:
            file.write(json_s)

    pass

    def rev_mode(self):
        self.is_converter = True
        for li in self.token_list:
            li.convert_type = TO_MUSIC
        self.rev_tokens: dict = {v: k for k, v in self.tokens.items()}
        print(self.rev_tokens)
        pass

    def begin_token(self, token_type):
        for i in range(len(self.rev_tokens)):
            if token_type in self.rev_tokens[i] :
                return i
        return "NOT FOUND"

    def end_token(self, token_type):
        for i in range(len(self.rev_tokens)):
            token = len(self.rev_tokens) - 1 - i
            if token_type in self.rev_tokens[token]:
                return token
